///
/// 9voice PSG for L3
///
/// @author Sasaji
/// @date   2016.05.25
///

module L3_9PSG
#( parameter dump_writes = 0 )
(
	input wire clk48,
	input wire sw_n,
	inout wire [7:0] data,
	input wire [3:0] addr,
	input wire cs_n,
	input wire we_n,
	input wire e,
	input wire clk2m,
	input wire res_p,
	output wire [2:0] led,
	output wire snd_out_0,
	output wire snd_out_1,
	output wire snd_out_2,
	output wire irq_p
);

wire res_n;
wire in_res_n;

wire clk_main;
wire clk_psg;
wire clk1m;
wire clk16;

// generate base clock
PLL	uPLL (
	.inclk0 ( clk48 ),
	.c0 ( clk_main ),
	.c1 ( clk16 )
);
//
assign clk_psg = clk2m;
assign clk1m = e;

// power on reset
wire por_n;
POR	uPOR
(
	.clk48	(clk48),
	.por_n	(por_n)
);
//
assign res_n = ~res_p;
assign in_res_n = (res_n & por_n & sw_n);

// VIA modoki
wire [7:0] pia_pia;
wire [7:0] pia_poa;
wire [7:0] pia_pib;
wire [7:0] pia_pob;
MY_VIA uMY_VIA
(
	.clk48	( clk48 ),		// fast clock
	.clk		( clk1m ),		// clock
	.res_n	( in_res_n ),	// reset
	.cs_n		( cs_n ),		// chip select
	.we_n		( we_n ),		// write enable
	.rs		( addr ),		// register select
	.da		( data ),		// data bus
	.pia		( pia_pia ),	// parallel input port A
	.poa		( pia_poa ),	// parallel output port A
	.pib		( pia_pib ),	// parallel input port B
	.pob		( pia_pob ),	// parallel output port B
	.irq		( irq_p )		// interrupt
);

// PSG Select
wire [3:0] psg_addr;
wire [7:0] psg_data_in;
wire psg_rd;
wire psg_we;
wire psg0_sel;
wire psg1_sel;
wire psg2_sel;
//
PSG_SEL uPSG_SEL (
	.bdir			( pia_pob[0] ),
	.bc1			( pia_pob[1] ),
	.psg0			( pia_pob[2] ),
	.psg1			( pia_pob[3] ),
	.psg2			( pia_pob[4] ),
	.data_in		( pia_poa ),
	.psg_rd		( psg_rd ),
	.psg_we		( psg_we ),
	.psg0_cs		( psg0_sel ),
	.psg1_cs		( psg1_sel ),
	.psg2_cs		( psg2_sel ),
	.addr_out	( psg_addr ),
	.data_out	( psg_data_in )
);

// PSG 0
wire [7:0] psg_data_out_0;
PSG #( dump_writes ) uPSG0 (
	.clk_main	( clk_main ),
	.clk_psg		( clk_psg ),
	.res_n		( in_res_n ),
	.cs			( psg0_sel ),
	.adr			( psg_addr ),
	.rd			( psg_rd ),
	.wr			( psg_we ),
	.data_in		( psg_data_in ),
	.data_out	( psg_data_out_0 ),
	.snd_out		( snd_out_0 )
);
// PSG 1
wire [7:0] psg_data_out_1;
PSG #( dump_writes ) uPSG1 (
	.clk_main	( clk_main ),
	.clk_psg		( clk_psg ),
	.res_n		( in_res_n ),
	.cs			( psg1_sel ),
	.adr			( psg_addr ),
	.rd			( psg_rd ),
	.wr			( psg_we ),
	.data_in		( psg_data_in ),
	.data_out	( psg_data_out_1 ),
	.snd_out		( snd_out_1 )
);
// PSG 2
wire [7:0] psg_data_out_2;
PSG #( dump_writes ) uPSG2 (
	.clk_main	( clk_main ),
	.clk_psg		( clk_psg ),
	.res_n		( in_res_n ),
	.cs			( psg2_sel ),
	.adr			( psg_addr ),
	.rd			( psg_rd ),
	.wr			( psg_we ),
	.data_in		( psg_data_in ),
	.data_out	( psg_data_out_2 ),
	.snd_out		( snd_out_2 )
);
// PSG Outdata Select
wire [7:0] psg_data_out;
assign pia_pia = psg_data_out;
//
PSG_OUTDATA_SEL uPSG_OUTDATA_SEL (
	.psg0_sel	( psg0_sel ),
	.psg1_sel	( psg1_sel ),
	.psg2_sel	( psg2_sel ),
	.psg0_data	( psg_data_out_0 ),
	.psg1_data	( psg_data_out_1 ),
	.psg2_data	( psg_data_out_2 ),
	.psg_data	( psg_data_out )
);

// red
assign led[0] = in_res_n;
// green
assign led[1] = 1'b1; 
// blue
assign led[2] = 1'b1;

endmodule

/// PSG Chip Select and Latch Address
module PSG_SEL
(
	input wire bdir,
	input wire bc1,
	input wire psg0,
	input wire psg1,
	input wire psg2,
	input wire [7:0] data_in,
	output wire psg_rd,
	output wire psg_we,
	output wire psg0_cs,
	output wire psg1_cs,
	output wire psg2_cs,
	output reg [3:0] addr_out,
	output reg [7:0] data_out
);
always @(bdir or bc1 or data_in) begin
	if (bdir) begin
		if (bc1) begin
			// latch address
			addr_out[3:0] = data_in[3:0];
		end
		else begin
			// write data
			data_out[7:0] = data_in[7:0];
		end
	end
end
//
assign psg_rd = (bc1 & ~bdir);
assign psg_we = bdir;
assign psg0_cs = (psg0 & (bc1 ^ bdir));
assign psg1_cs = (psg1 & (bc1 ^ bdir));
assign psg2_cs = (psg2 & (bc1 ^ bdir));

endmodule

/// PSG outdata select
module PSG_OUTDATA_SEL
(
	input wire psg0_sel,
	input wire psg1_sel,
	input wire psg2_sel,
	input wire [7:0] psg0_data,
	input wire [7:0] psg1_data,
	input wire [7:0] psg2_data,
	output wire [7:0] psg_data
);
//
function [7:0] psg_data_sel;
	input psg0_sel;
	input psg1_sel;
	input psg2_sel;
	input [7:0] psg0_data;
	input [7:0] psg1_data;
	input [7:0] psg2_data;
begin
	if (psg0_sel) psg_data_sel = psg0_data;
	else if (psg1_sel) psg_data_sel = psg1_data;
	else if (psg2_sel) psg_data_sel = psg2_data;
	else psg_data_sel = psg0_data;
end
endfunction
//
assign psg_data = psg_data_sel(psg0_sel, psg1_sel, psg2_sel,
	psg0_data, psg0_data, psg2_data);

endmodule

///
/// PSG Unit
///
module PSG
#( parameter dump_writes = 0 )
(
	input wire clk_main,
	input wire clk_psg,
	input wire res_n,
	input wire cs,
	input wire [3:0] adr,
	input wire rd,
	input wire wr,
	input wire [7:0] data_in,
	output wire [7:0] data_out,
	output wire snd_out
);

wire [3:0] out4_a;
wire [3:0] out4_b;
wire [3:0] out4_c;
SQMUSIC #( dump_writes ) uSQMUSIC
(
	.reset_n		( res_n ),
	.clk			( clk_psg ), // normally slower than the CPU clock
	.data_in		( data_in ),
	.data_out	( data_out ),
	.adr			( adr ),
	.rd			( cs & rd ), // read
	.wr			( cs & wr ), // write
	.A				( out4_a ),
	.B				( out4_b ),
	.C				( out4_c ) // channel outputs
);

//
wire [5:0] out6_a;
VOLCONV #( dump_writes ) uVOLCONV_A (
	.invol		( out4_a ),
	.outvol		( out6_a )
);
wire [5:0] out6_b;
VOLCONV #( dump_writes ) uVOLCONV_B (
	.invol		( out4_b ),
	.outvol		( out6_b )
);
wire [5:0] out6_c;
VOLCONV #( dump_writes ) uVOLCONV_C (
	.invol		( out4_c ),
	.outvol		( out6_c )
);

//
wire [7:0] out8_all;
assign out8_all = out6_a + out6_b + out6_c;

// PWM
wire out_pwm;
MY_PWM1 #( 73 ) uMY_PWM1 (
	.clk			( clk_main ),
	.res_n		( res_n ),
	.in			( out8_all ),
	.out			( snd_out )
);

endmodule

// volume convert
module VOLCONV
#( parameter dump_writes = 0 )
(
	input wire [3:0] invol,
	output reg [5:0] outvol
);

always @(invol) begin
	case(invol)
	4'd15: outvol <= 6'd24;
	4'd14: outvol <= 6'd20;
	4'd13: outvol <= 6'd17;
	4'd12: outvol <= 6'd15;
	4'd11: outvol <= 6'd13;
	4'd10: outvol <= 6'd11;
	default: outvol <= { 2'b0, invol };
	endcase
/*
	case(invol)
	4'd15: outvol <= 6'd63;
	4'd14: outvol <= 6'd49;
	4'd13: outvol <= 6'd38;
	4'd12: outvol <= 6'd30;
	4'd11: outvol <= 6'd22;
	4'd10: outvol <= 6'd17;
	4'd9: outvol <= 6'd13;
	4'd8: outvol <= 6'd10;
	4'd7: outvol <= 6'd8;
	default: outvol <= { 2'b0, invol };
	endcase
*/
end

endmodule

